import numpy as np
import torch
from pandas import DataFrame

from autofaq.clean.cleaner import Cleaner
from autofaq.config.configurable import Configurable
from autofaq.util.out import sprint


class SubjectCleaner(Cleaner, Configurable):
    def clean(self, df: DataFrame, aux: DataFrame = None):
        sprint("Filtering data using subject cleaner ...", fg="cyan")
        embeddings = torch.load(".cache/embeddings.bin")
        embeddings = {k: v for k, v in embeddings.items() if isinstance(k, int)}
        q_embeds = torch.cat(tuple(v["q"].reshape(1, -1) for v in embeddings.values()))
        q_mean = torch.mean(q_embeds, axis=0, keepdim=True)
        distances = torch.cdist(q_embeds, q_mean).reshape(-1)
        m, std = distances.mean(), distances.std()
        m, std = m.item(), std.item()
        sprint(
            "Distance from subject centeroid => mean: @m, std: @std",
            fg="black",
            m=(m, "magenta"),
            std=(std, "magenta"),
        )
        threshold = m + self.config.distance_std * std
        filter_ = {k: bool(v < threshold) for k, v in zip(embeddings.keys(), distances)}
        return df.apply(lambda x: filter_[x["id"]], axis=1)

    def settings(self):
        return {
            "namespace": "clean.subject",
            "distance_std": (float, 1, False, ""),
        }



from os import path

import numpy as np
import torch
from onnxruntime import ExecutionMode, InferenceSession, SessionOptions
from transformers import AutoTokenizer

# https://medium.com/microsoftazure/accelerate-your-nlp-pipelines-using-hugging-face-transformers-and-onnx-runtime-2443578f4333
# https://huggingface.co/transformers/serialization.html
# https://huggingface.co/sentence-transformers/paraphrase-xlm-r-multilingual-v1


# Mean Pooling - Take attention mask into account for correct averaging
def mean_pooling(token_embeddings, attention_mask):
    input_mask_expanded = (
        attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
    )
    return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(
        input_mask_expanded.sum(1), min=1e-9
    )


def openXLMSession(modelPath):
    tokenizer = AutoTokenizer.from_pretrained(modelPath, use_fast=True)
    options = SessionOptions()
    options.intra_op_num_threads = 1
    options.execution_mode = ExecutionMode.ORT_PARALLEL
    session = InferenceSession(
        path.join(modelPath, "xlm.onnx"),
        options,
        providers=[
            "TensorrtExecutionProvider",
            "CUDAExecutionProvider",
            "CPUExecutionProvider",
        ],
    )
    return {"session": session, "tokenizer": tokenizer}


def embedSentences(xlmSession, sentences):
    encoded_input = xlmSession["tokenizer"](
        sentences, padding=True, truncation=True, return_tensors="pt"
    )
    tokens = {name: np.atleast_2d(val) for name, val in encoded_input.items()}
    out1, out2 = xlmSession["session"].run(None, tokens)
    res = mean_pooling(torch.tensor(out1), encoded_input["attention_mask"])
    return res



from autofaq.config.configurable import Configurable
from autofaq.language_model.embedder import BaseEmbedder
from autofaq.language_model.xlm import embedSentences, openXLMSession


class XLMEmbedder(BaseEmbedder, Configurable):
    def settings(self):
        return {
            "namespace": "embed.xlm",
            "model_path": (str, "MODEL_PATH", True, "xlm-r model path (onnx)"),
        }

    def on_bound(self):
        self.session = openXLMSession(self.config.model_path)

    def embed(self, sentences: list[str]):
        res = embedSentences(self.session, sentences)
        return res



from autofaq.cli.commands.clean import clean
from autofaq.cli.commands.embed import embed
from autofaq.cli.commands.extract import extract
from autofaq.cli.commands.init import init
from autofaq.cli.commands.keywords import keywords
from autofaq.cli.commands.render import render
from autofaq.cli.commands.search import search



import numpy as np
import torch
from pandas import DataFrame

from autofaq.clean.cleaner import Cleaner
from autofaq.config.configurable import Configurable
from autofaq.util.out import sprint


class EntailmentCleaner(Cleaner, Configurable):
    def clean(self, df: DataFrame, aux: DataFrame = None):
        sprint("Filtering data using entailment cleaner ...", fg="cyan")
        embeddings = torch.load(".cache/embeddings.bin")
        embeddings = {k: v for k, v in embeddings.items() if isinstance(k, int)}
        scores_map = {k: torch.dot(v["q"], v["a"]) for k, v in embeddings.items()}
        self_scores = {k: torch.dot(v["q"], v["q"]) for k, v in embeddings.items()}
        deviation_map = {k: np.abs(scores_map[k] - self_scores[k]) for k in scores_map}
        ids = list(df["id"])
        scores_map = {k: v for k, v in scores_map.items() if k in ids}
        scores = list(scores_map.values())
        m, var = np.mean(scores), np.std(scores)
        sprint(
            "Entailment scores => mean: @m, std: @std",
            fg="black",
            m=(m, "magenta"),
            std=(var, "magenta"),
        )
        deviations = list(deviation_map.values())
        m_dev, std_dev = np.mean(deviations), np.std(deviations)
        sprint(
            "QA Deviation scores => mean: @m, std: @std",
            fg="black",
            m=(m_dev, "magenta"),
            std=(std_dev, "magenta"),
        )
        aux["entailment-score"] = df.apply(lambda x: scores_map[x["id"]], axis=1)
        aux["entailment-deviation"] = df.apply(lambda x: deviation_map[x["id"]], axis=1)
        r = df.apply(
            lambda x: bool(
                (scores_map[x["id"]] > (m + self.config.entailment_std * var))
                and (
                    (not self.config.deviation_filter)
                    or (
                        (
                            deviation_map[x["id"]]
                            > (m_dev - self.config.deviation_std * std_dev)
                        )
                        and (
                            deviation_map[x["id"]]
                            < (m_dev + self.config.deviation_std * std_dev)
                        )
                    )
                )
            ),
            axis=1,
        )
        return r

    def settings(self):
        return {
            "namespace": "clean.entailment",
            "entailment_std": (float, 0, False, ""),
            "deviation_filter": (bool, True, False, ""),
            "deviation_std": (float, 2, False, ""),
        }



import json
import pickle
import subprocess
from os.path import exists as fexists
from os.path import join as pjoin

import click
import numpy as np
import pandas as pd
import requests
import torch
from tqdm import tqdm
from user_agent import generate_user_agent

from autofaq.cli.entry import entry
from autofaq.config.binder import bind
from autofaq.language_model.xlm_embedder import XLMEmbedder
from autofaq.util.hash import sha256id
from autofaq.util.out import sprint

embedders = {
    "xlm": XLMEmbedder,
}


@entry.command(help="Computes vector embeddings of the dataset")
@click.option(
    "-s", "--scratch", default=False, is_flag=True, help="ignore present embeddings"
)
@click.option(
    "-e",
    "--embedder",
    default="xlm",
    help="embedder module",
    type=click.Choice(list(embedders.keys())),
)
@click.argument("filter_name", required=False)
def embed(filter_name, embedder, scratch):
    sprint("Beginning the embedding process for project ...", fg="cyan")

    df = pd.read_csv("dataset.csv")
    if filter_name is not None:
        filter_p = f"{filter_name}.filter"
        with open(filter_p, "rb") as f:
            filter_i = pickle.load(f)
        selection = df.apply(lambda x: filter_i[x["id"]], axis=1)
        df = df[selection]
    o_df = df.reset_index(drop=True)

    sprint("Initiating ONNX session for language model ...", fg="black")
    embedder = embedders[embedder]()
    ok = bind(embedder, "settings.toml")
    if not ok:
        return

    e_path = ".cache/embeddings.bin"
    if fexists(e_path) and not scratch:
        embeddings = torch.load(e_path)
    else:
        embeddings = {}

    # Filter out existants here
    df = o_df[o_df["id"].apply(lambda x: x not in embeddings)]
    df.reset_index(drop=True)

    sprint("Embedding QA pairs ...", fg="black")
    sentences = list(df.q) + list(df.a)
    batches = batch(sentences)
    s_embeds = embed_batches(embedder, batches)

    n = len(df)

    for i, row in df.iterrows():
        q_e = s_embeds[i]
        a_e = s_embeds[n + i]
        embeddings[row["id"]] = {
            "q": q_e,
            "a": a_e,
        }

    df = o_df[o_df.src_title.apply(lambda x: f"t-{sha256id(x)}" not in embeddings)]
    df.reset_index(drop=True)

    sprint("Embedding page titles ...", fg="black")
    titles = list(set(df.src_title.apply(lambda x: x.strip())))
    batches = batch(titles)
    t_embeds = embed_batches(embedder, batches)
    for title, embedding in zip(titles, t_embeds):
        embeddings[f"t-{sha256id(title)}"] = embedding

    torch.save(embeddings, e_path)
    sprint("Successfully calculated and saved the embeddings!", fg="green")


def batch(items, size=16):
    for i in range(0, len(items), size):
        yield items[i : i + size]


def embed_batches(embedder, batches):
    embeddings = ()
    batches = list(batches)  # To track progress
    for b in tqdm(batches):
        res = embedder.embed(b)
        if len(embeddings) == 0:
            embeddings = res
        else:
            embeddings = torch.cat((embeddings, res))
    return embeddings



from autofaq.render.renderer import BaseRenderer


class MarkdownRenderer(BaseRenderer):
    def render(self, data: list[dict], out: str):
        document = []
        for page in data:
            title = page["title"]
            document.append(f"### {title}")
            for pair in page["pairs"]:
                q = pair["q"]
                a = pair["a"]
                aux_data = pair["aux"]
                aux_data["tag"] = page["query"]

                document.append(f"#### {q}")
                document.append(a)
                aux_data = {
                    k: v if not isinstance(v, float) else f"{v:.2f}"
                    for k, v in aux_data.items()
                }
                aux_data = ", ".join(f"{k}={v}" for k, v in aux_data.items())
                if aux_data.strip() != "":
                    document.append(f"\n```{aux_data}```")

        doc = "\n".join(document)
        pt = f"{out}.md"
        with open(pt, "w") as f:
            f.write(doc)
        return pt



from bs4 import BeautifulSoup
from bs4.element import Comment

from autofaq.extract.extractor import BaseExtractor


class EasyExtractor(BaseExtractor):
    def extract(self, content: str) -> list[dict]:
        qa = self.retreive_qa(content)
        return [{"q": q, "a": " ".join(a)} for q, a in qa]

    def parent_sig(self, p):
        tags = []
        t = p
        while t is not None:
            tags.insert(0, t.name)
            t = t.parent
        cls = "|".join(sorted(p.attrs.get("class", [])))
        hi = ">".join(tags)
        return hi + "+" + cls

    def tag_visible(self, element):
        if element.parent.name in [
            "style",
            "script",
            "head",
            "title",
            "meta",
            "[document]",
        ]:
            return False
        if isinstance(element, Comment):
            return False
        return True

    def text_from_html(self, body):
        soup = BeautifulSoup(body, "html.parser")
        soup = soup.find("body")
        texts = soup.findAll(text=True)
        visible_texts = filter(self.tag_visible, texts)
        r = [(t.strip(), self.parent_sig(t.parent)) for t in visible_texts]
        return list(filter(lambda x: x[0] != "", r))

    def is_question(self, sentence: str) -> bool:
        # This supports persian and english questions
        # TODO: extend it to much better one
        return sentence.strip().endswith("؟") or sentence.strip().endswith("?")

    def retreive_qa(self, html_content):
        q_list = self.text_from_html(html_content)
        stack = []
        cache = []
        current = None
        last = None
        for s, id_ in q_list:
            if self.is_question(s):
                if current:
                    stack.append((current[0], cache))
                    cache = []
                    last = None
                current = s, id_
            else:
                if current:
                    if last is None or last == id_:
                        cache.append(s)
                        last = id_
        return stack



import json
import subprocess
from concurrent.futures import ThreadPoolExecutor, as_completed
from hashlib import sha256
from os.path import exists as fexists
from os.path import join as pjoin
from urllib.parse import unquote

import click
import pandas as pd
import requests
from tqdm import tqdm
from user_agent import generate_user_agent

from autofaq.cli.entry import entry
from autofaq.extract.better_extractor import BetterExtractor
from autofaq.extract.easy_extractor import EasyExtractor
from autofaq.util.out import sprint

headers = {
    "User-Agent": generate_user_agent(),
    "Accept-Language": "en-US,en;q=0.9,fa-IR;q=0.8,fa;q=0.7",
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
}

extractors = {
    "easy": EasyExtractor,
    "better": BetterExtractor,
}


def fetch_content(url, retry=False):
    hash_ = sha256(url.encode("utf-8")).hexdigest()
    tg = f".cache/{hash_}"
    try:
        if fexists(tg):
            with open(tg, "rb") as f:
                c = f.read()
                if c == "|error|" and not retry:
                    return hash_, False
                return hash_, True, c
        response = requests.get(url, timeout=4)
        if response.status_code == 200:
            with open(tg, "wb") as f:
                f.write(response.content)
            return hash_, True, response.content
        else:
            with open(tg, "wb") as f:
                f.write("|error|".encode("utf-8"))
            return hash_, False, None
    except requests.exceptions.RequestException:
        with open(tg, "wb") as f:
            f.write("|error|".encode("utf-8"))
        return hash_, False, None


def download_contents(urls, num_threads=8):
    contents = {}
    with ThreadPoolExecutor(max_workers=num_threads) as executor:
        futures = {executor.submit(fetch_content, url): url for url in urls}

        for future in tqdm(as_completed(futures), total=len(urls)):
            hash_, ok, content = future.result()
            contents[hash_] = (ok, content)

    return contents


@entry.command(help="Collects webpages and extracts QA pairs")
@click.option(
    "-e",
    "--extractor",
    default="easy",
    help="extractor engine",
    type=click.Choice(list(extractors.keys())),
)
def extract(extractor):
    sprint("Starting the extraction process ...", fg="cyan")

    pages = pd.read_csv("search.csv")
    r = pages.to_dict(orient="index")
    r = list(r.values())

    sprint("Gathering webpages from web:", fg="black")
    errors = []
    content_map = download_contents(list(map(lambda x: x["link"], r)))
    for d in r:
        hash_ = sha256(d["link"].encode("utf-8")).hexdigest()
        ok, content = content_map[hash_]
        if ok:
            d["state"] = 200
            d["html"] = content
        else:
            errors.append(d["link"])
            d["state"] = -1
            d["html"] = ""

    if len(errors) != 0:
        sprint(
            "Couldn't gather @count webpages! List:",
            fg="red",
            count=(len(errors), "red", "bold"),
        )
        for e in errors:
            sprint(unquote(e), fg="black")

    sprint("Extracting QA pairs ...", fg="black")

    extractor = extractors[extractor]()
    for d in tqdm(r):
        try:
            qa = extractor.extract(d["html"])
            d["qa"] = qa
        except Exception:
            d["qa"] = []

    dataset = []

    for d in tqdm(r):
        src_title = d["title"]
        src_link = d["link"]
        qa = d["qa"]
        for pair in qa:
            dataset.append(
                {
                    "src_title": src_title,
                    "src_link": src_link,
                    **pair,
                }
            )
    data = pd.DataFrame(dataset)
    data["id"] = range(len(data))
    data.to_csv("dataset.csv", index=False)
    sprint("Successfully saved QA pairs to dataset.csv!", fg="green")



import re

from click import echo, style


def sprint(msg, fg=None, **kwargs):
    # Structure -> Here is message @msg_number
    pattern = "(@[A-Za-z0-9_]+)"
    msg_split = re.split(pattern, msg)
    styled = {}
    for k, v in kwargs.items():
        match v:
            case (value, color, st):
                styled[f"@{k}"] = style(value, fg=color, bold=st == "bold")
            case (value, color):
                styled[f"@{k}"] = style(value, fg=color)
            case value:
                styled[f"@{k}"] = f"{value}"
    new_msg = []
    for m in msg_split:
        if m in styled:
            new_msg.append(styled[m])
        else:
            if fg:
                new_msg.append(style(m, fg=fg))
            else:
                new_msg.append(m)
    print("".join(new_msg))



import numpy as np
import torch
from pandas import DataFrame

from autofaq.clean.cleaner import Cleaner
from autofaq.config.configurable import Configurable
from autofaq.util.hash import sha256id
from autofaq.util.out import sprint

blacklist = {
    "fa": [
        "نمونه سوال",
    ]
}


def filter_title(filter_map, title):
    title = title.strip()
    q = filter_map[sha256id(title)]
    if q:
        for lang, l_ in blacklist.items():
            for w in l_:
                if w in title:
                    return False
    return q


class PageCleaner(Cleaner, Configurable):
    def clean(self, df: DataFrame, aux: DataFrame = None):
        sprint("Filtering data using page title cleaner ...", fg="cyan")
        embeddings = torch.load(".cache/embeddings.bin")
        titles = list(set(df.src_title.apply(lambda x: x.strip())))

        title_map = {}
        for t in titles:
            s = sha256id(t)
            title_map[s] = embeddings[f"t-{s}"]

        t_embeds = torch.cat(tuple(v.reshape(1, -1) for v in title_map.values()))
        q_mean = torch.mean(t_embeds, axis=0, keepdim=True)
        distances = torch.cdist(t_embeds, q_mean).reshape(-1)
        m, std = distances.mean(), distances.std()
        m, std = m.item(), std.item()
        sprint(
            "Distance from title centeroid => mean: @m, std: @std",
            fg="black",
            m=(m, "magenta"),
            std=(std, "magenta"),
        )
        threshold = m + self.config.distance_std * std
        filter_ = {k: bool(v < threshold) for k, v in zip(title_map.keys(), distances)}
        return df.apply(lambda x: filter_title(filter_, x["src_title"]), axis=1)

    def settings(self):
        return {
            "namespace": "clean.page",
            "distance_std": (float, 1, False, ""),
        }



import json
import subprocess
from os.path import exists as fexists
from os.path import join as pjoin

import click
import pandas as pd

from autofaq.cli.entry import entry
from autofaq.config.binder import bind
from autofaq.search.ddg_engine import DDGEngine
from autofaq.search.google_serp_search import SerpEngine
from autofaq.search.xng_engine import XNGEngine
from autofaq.util.out import sprint

engines = {"ddg": DDGEngine, "google": SerpEngine, "xng": XNGEngine}
langs = {"fa": None, "en": None}


@entry.command(help="Compiles search results for keywords")
@click.option(
    "-e",
    "--engine",
    default="ddg",
    help="Search Engine",
    type=click.Choice(engines.keys()),
)
def search(engine):
    sprint("Gathering search results for keywords ...", fg="cyan")
    query_list = pd.read_csv("keywords.csv")
    query_list = query_list[query_list["include"]]
    query_list = list(query_list["query"])

    engine = engines[engine]()
    ok = bind(engine, "settings.toml")
    if not ok:
        return

    results = engine.search(query_list)
    results.to_csv("search.csv", index=False)
    sprint("Successfully saved results to search.csv!", fg="green")



from abc import ABC, abstractmethod
from types import SimpleNamespace

from autofaq.util.out import sprint


class Configurable(ABC):
    def __init__(self):
        self.config = SimpleNamespace()

    @abstractmethod
    def settings(self) -> dict:
        """
        This method should return a dict showing the config of this object
        {
            "namespace": "namespace",
            "var": (type, default, required, hint)
        }
        """
        pass

    def on_bound(self):
        pass

    def configure(self, config: dict):
        settings = self.settings()
        namespace = settings.pop("namespace", "")
        for k, v in settings.items():
            if namespace:
                k = f"{namespace}.{k}"
            type_, def_, req_, hint_ = v
            ok = self.read_key(config, k, default=def_, required=req_, type_=type_)
            if not ok:
                return False
        self.on_bound()
        return True

    @classmethod
    def _nested_get(cls, d: dict, path: list):
        r = d
        path = path.copy()
        while len(path) != 0:
            k = path.pop(0)
            if k in r:
                r = r[k]
            else:
                return None
        return r

    def read_key(
        self, config: dict, key: str, default=None, required=False, type_=None
    ):
        path = key.split(".")
        value = self._nested_get(config, path)
        if value is None and required:
            sprint(
                "You must provide a value for @key key in your config!",
                fg="red",
                key=(key, "white", "bold"),
            )
            return False
        if required and value == default:
            sprint(
                "You must change the default value for @key key in your config!",
                fg="red",
                key=(key, "white", "bold"),
            )
            return False
        if value is None:
            value = default
        else:
            if type_:
                value = type_(value)
        setattr(self.config, path[-1], value)
        return True



import json
import pickle
import subprocess
from hashlib import sha256
from os.path import exists as fexists
from os.path import join as pjoin

import click
import numpy as np
import pandas as pd
import requests
from tqdm import tqdm
from user_agent import generate_user_agent

from autofaq.clean.classic import ClassicCleaner
from autofaq.clean.entailment import EntailmentCleaner
from autofaq.clean.fuzzy_match import FuzzCleaner
from autofaq.clean.page import PageCleaner
from autofaq.clean.subject import SubjectCleaner
from autofaq.cli.entry import entry
from autofaq.config.binder import bind
from autofaq.util.out import sprint

cleaners = {
    "classic": ClassicCleaner,
    "entailment": EntailmentCleaner,
    "subject": SubjectCleaner,
    "fuzz": FuzzCleaner,
    "page": PageCleaner,
}


@entry.command(help="Refines the dataset")
@click.option(
    "-c",
    "--cleaner",
    default="classic",
    help="cleaner module",
    type=click.Choice(list(cleaners.keys())),
)
@click.option(
    "-s", "--scratch", default=False, is_flag=True, help="use existing filter"
)
@click.option(
    "-i", "--inplace", default=False, is_flag=True, help="use existing filter"
)
@click.argument("name", required=False)
def clean(name, cleaner, scratch, inplace):
    df = pd.read_csv("dataset.csv")
    if name is None and not inplace:
        sprint("You must provide a filter name when in not inplace mode!", fg="red")
        return
    filter_p = f"{name}.filter"
    filter_a = f"{name}.aux"
    filter_i = {}
    if fexists(filter_p) and not scratch:
        with open(filter_p, "rb") as f:
            filter_i = pickle.load(f)
        selection = df.apply(lambda x: filter_i[x["id"]], axis=1)
        df = df[selection]

    aux = pd.DataFrame()
    aux["id"] = df["id"]
    if fexists(filter_a) and not scratch:
        aux = pd.read_pickle(filter_a)

    cleaner = cleaners[cleaner]()
    ok = bind(cleaner, "settings.toml")
    if not ok:
        return

    selection = cleaner.clean(df, aux=aux)
    sprint(
        "Survivors: @s @sep @all",
        fg="white",
        s=(selection.sum(), "green", "bold"),
        all=(len(selection), "black"),
        sep=("/", "black"),
    )
    if inplace:
        df[selection].to_csv("dataset.csv", index=False)
    else:
        filter_ = pd.DataFrame()
        filter_["id"] = df["id"]
        filter_["selected"] = selection
        filter_d = {x["id"]: x["selected"] for _, x in filter_.iterrows()}
        filter_d = {**filter_i, **filter_d}
        aux.to_pickle(f"{name}.aux")
        with open(filter_p, "wb") as f:
            pickle.dump(filter_d, f)



import numpy as np
import torch
from fuzzywuzzy.fuzz import ratio
from pandas import DataFrame

from autofaq.clean.cleaner import Cleaner
from autofaq.config.configurable import Configurable
from autofaq.util.out import sprint


class FuzzCleaner(Cleaner, Configurable):
    def clean(self, df: DataFrame, aux: DataFrame = None):
        sprint("Filtering data using fuzzy cleaner ...", fg="cyan")
        aux["fuzz"] = df.apply(lambda x: ratio(x.q, x.a), axis=1)
        m, var = aux["fuzz"].mean(), aux["fuzz"].std()
        sprint(
            "QA fuzz ratio => mean: @m, std: @std",
            fg="black",
            m=(m, "magenta"),
            std=(var, "magenta"),
        )
        threshold = m + self.config.fuzz_std * var
        return df.apply(lambda x: ratio(x.q, x.a) < threshold, axis=1)

    def settings(self):
        return {
            "namespace": "clean.fuzz",
            "fuzz_std": (float, 1, False, ""),
        }



from tomlkit import parse

from autofaq.config.configurable import Configurable


def bind(obj, config: str | dict):
    if isinstance(config, str):
        with open(config, "r") as f:
            config = parse(f.read())
    if isinstance(obj, Configurable):
        return obj.configure(config)
    return True



import subprocess
from os.path import join as pjoin

import click
import tomlkit
from tomlkit import boolean, document, nl, table

from autofaq.clean.entailment import EntailmentCleaner
from autofaq.clean.fuzzy_match import FuzzCleaner
from autofaq.clean.page import PageCleaner
from autofaq.clean.subject import SubjectCleaner
from autofaq.cli.entry import entry
from autofaq.language_model.xlm_embedder import XLMEmbedder
from autofaq.search.ddg_engine import DDGEngine
from autofaq.search.google_serp_search import SerpEngine
from autofaq.search.xng_engine import XNGEngine
from autofaq.util.dir import create_directories_and_settings
from autofaq.util.out import sprint

configurables = [
    EntailmentCleaner,
    FuzzCleaner,
    SubjectCleaner,
    PageCleaner,
    DDGEngine,
    SerpEngine,
    XNGEngine,
    XLMEmbedder,
]


@entry.command(help="Initiates a new mining project")
@click.option("-p", "--path", default=".", help="base path")
@click.option(
    "-f", "--force", default=False, is_flag=True, help="overwrite project if exists"
)
@click.argument("name")
def init(name, path, force):
    sprint("Initializing new mining project ...", fg="cyan")

    structure = {
        ".cache": None,
    }

    doc = document()
    doc.add("name", name)
    doc.add(nl())
    configs = gather_configs(doc)

    ok = create_directories_and_settings(structure, configs, pjoin(path, name), force)
    if ok:
        sprint("Successfully created project!", fg="green")
    else:
        sprint("Error creating project!", fg="red")


def gather_configs(d=None):
    all_ = {}
    for c in configurables:
        o = c()
        s = o.settings()
        namespace = s.pop("namespace", None)
        for k, v in s.items():
            if namespace:
                k = f"{namespace}.{k}"
            all_[k] = v
    if d is None:
        d = document()
    for k, v in all_.items():
        path = k.split(".")
        *trail, last = path
        cur = d
        for t in trail:
            if t not in cur:
                cur.add(t, table())
            cur = cur[t]
        type_, def_, req_, hint = v
        cm = f"{hint} (type: {type_.__name__})".strip()
        if type_ == bool:
            item = boolean("true" if def_ else "")
            item = item.comment(cm)
            cur.add(last, item)
        else:
            cur.add(last, def_)
            cur[last].comment(cm)
    return d



import os

from tomlkit import TOMLDocument, dump


def create_directories(dir_structure: dict, start_path: str = "."):
    for directory, subdirectories in dir_structure.items():
        os.makedirs(os.path.join(start_path, directory), exist_ok=True)
        if isinstance(subdirectories, dict):
            create_directories(subdirectories, os.path.join(start_path, directory))
        elif isinstance(subdirectories, set):
            for subdirectory in subdirectories:
                os.makedirs(
                    os.path.join(start_path, directory, subdirectory), exist_ok=True
                )


def create_directories_and_settings(
    directory_structure: dict,
    settings: TOMLDocument,
    start_path: str = ".",
    force=False,
):
    if os.path.exists(start_path) and os.listdir(start_path) is not None and not force:
        return False
    create_directories(directory_structure, start_path)
    with open(os.path.join(start_path, "settings.toml"), "w") as f:
        dump(settings, f)
    return True



import subprocess
from os.path import exists as fexists
from os.path import join as pjoin

import click
import pandas as pd

from autofaq.cli.entry import entry
from autofaq.search.google_keywords import GoogleKeywords
from autofaq.util.out import sprint

modes = ["single", "google", "general"]

langs = {"fa": ("سوالات متداول", ""), "en": ("frequently asked questions", "in")}


@entry.command(help="Generates an expanded list of keywords")
@click.option(
    "-a", "--append", default=True, is_flag=True, help="append if keyword list exists"
)
@click.option("-d", "--digits", default=False, is_flag=True, help="use digits in hints")
@click.option(
    "-m",
    "--mode",
    default="google",
    help="mode for creating keywords",
    type=click.Choice(modes),
)
@click.option(
    "-l",
    "--language",
    default="fa",
    help="target language",
    type=click.Choice(list(langs.keys())),
)
@click.argument("keyword", required=False)
def keywords(keyword, append, digits, mode, language):
    sprint("Gathering relevant keywords for project ...", fg="cyan")
    if mode == "single":
        data = [keyword]
    elif mode == "google" or mode == "general":
        d_kw = "-nd" if digits else ""
        kw = GoogleKeywords(lang=f"{language}{d_kw}")
        f_word, f_ap = langs[language]
        m_kw = keyword if mode == "google" else ""
        if mode == "google" and keyword is None:
            sprint("Keyword required for google mode!", fg="red")
            return

        df = kw.load(f"{f_word} {f_ap} {m_kw}".replace("  ", " ").strip())
        data = list(df.suggestion)
    if append and fexists("keywords.csv"):
        prev_data = list(pd.read_csv("keywords.csv")["query"])
        data = [*prev_data, *data]
    df = pd.DataFrame()
    df["id"] = range(len(data))
    df["query"] = data
    df["include"] = True
    df.to_csv("keywords.csv", index=False)
    sprint("Successfully saved keywords to keywords.csv!", fg="green")



import json
from json import JSONEncoder

import torch

from autofaq.render.renderer import BaseRenderer


class TEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, torch.Tensor):
            return o.item()
        return o.__dict__


class JSONRenderer(BaseRenderer):
    def render(self, data: list[dict], out: str):
        pt = f"{out}.json"
        with open(pt, "w") as f:
            json.dump(data, f, ensure_ascii=False, indent=4, cls=TEncoder)
        return pt



from setuptools import find_packages, setup

setup(
    name="auto-faq",
    version="0.4.0",
    description="Automatic FAQ Mining Framework",
    license="MIT",
    packages=find_packages(exclude=["test"]),
    author="Amin Rezaei",
    author_email="AminRezaei0x443@gmail.com",
    keywords=[],
    entry_points={
        "console_scripts": [
            "autofaq = autofaq.cli.entry:entry",
        ],
    },
    url="https://github.com/AminRezaei0x443/auto-faq",
    install_requires=[],
    extras_require={},
)



from pandas import DataFrame

from autofaq.clean.cleaner import Cleaner
from autofaq.util.out import sprint


class ClassicCleaner(Cleaner):
    def clean(self, df: DataFrame, **kwargs):
        sprint("Filtering data using classic cleaner ...", fg="cyan")
        selection = df.a.apply(lambda x: type(x) == str and x.strip() != "")
        return selection



import json
import pickle
import re
import subprocess
from hashlib import sha256
from os.path import exists as fexists
from os.path import join as pjoin

import click
import numpy as np
import pandas as pd
import requests
import torch
from tqdm import tqdm
from user_agent import generate_user_agent

from autofaq.cli.entry import entry
from autofaq.language_model.xlm import embedSentences, openXLMSession
from autofaq.render.json_renderer import JSONRenderer
from autofaq.render.md_renderer import MarkdownRenderer
from autofaq.util.out import sprint

# This matches languages characters of persian and english (no digits)
# Aim: Easy clearance of numbered lists and questions
alpha_L = r"[^\u0041-\u005A\u0061-\u007A\u0622\u0626-\u0628\u062a-\u063a\u0641\u0642\u0644-\u0648\u067e\u0686\u0698\u06a9\u06af\u06cc\u200f\u0020]+"

renderers = {
    "md": MarkdownRenderer,
    "markdown": MarkdownRenderer,
    "json": JSONRenderer,
}


def replace_prefix_pattern(target, pattern):
    r = re.search(pattern, target, re.UNICODE)
    if r is None:
        return target
    begin_i, end_i = r.span()
    if begin_i == 0:
        return target[end_i:]
    return target


@entry.command(help="Transforms the dataset into human-readable formats")
@click.option(
    "-r",
    "--renderer",
    default="md",
    help="renderer engine",
    type=click.Choice(list(renderers.keys())),
)
@click.argument("filter_name", required=False)
def render(filter_name, renderer):
    sprint("Rendering QA pairs to markdown ...", fg="cyan")

    df = pd.read_csv("dataset.csv")
    aux = df[["id"]]

    if filter_name is not None:
        filter_p = f"{filter_name}.filter"
        filter_a = f"{filter_name}.aux"
        with open(filter_p, "rb") as f:
            filter_i = pickle.load(f)
        aux = pd.read_pickle(filter_a)
        selection = df.apply(lambda x: filter_i[x["id"]], axis=1)
        selection_aux = aux.apply(lambda x: filter_i[x["id"]], axis=1)

        df = df[selection]
        aux = aux[selection_aux]

    aux = aux.set_index("id")
    df = df.reset_index(drop=True)

    # df -> src_title, src_link, q, a
    pages = {row.src_link: row.src_title for _, row in df.iterrows()}
    grouped = {
        link: list(df[df.src_link == link].apply(lambda x: (x["id"], x.q, x.a), axis=1))
        for link in pages
    }

    query_df = pd.read_csv("search.csv")
    query_df = query_df.set_index("link", drop=True)

    data = []
    for group, pairs in grouped.items():
        title = pages[group]
        query = query_df.loc[group]["q"]
        if not isinstance(query, str):
            query = query.iloc[0]
        page = {
            "title": title,
            "query": query,
            "pairs": [],
        }
        for id_, q, a in pairs:
            q = q.replace("\n", "").strip()
            q = replace_prefix_pattern(q, alpha_L)

            aux_data = dict(aux.loc[id_])
            page["pairs"].append(
                {
                    "q": q,
                    "a": a,
                    "aux": aux_data,
                }
            )
        data.append(page)

    renderer = renderers[renderer]()
    o = renderer.render(data, "out")

    sprint("Successfully rendered data to @o!", fg="green", o=(o, "green", "bold"))
